product_map = {   1: 'Original 1000',
    3: 'Color 650',
    10: 'White 800 (Low Voltage)',
    11: 'White 800 (High Voltage)',
    18: 'White 900 BR30 (Low Voltage)',
    20: 'Color 1000 BR30',
    22: 'Color 1000',
    27: 'LIFX A19',
    28: 'LIFX BR30',
    29: 'LIFX+ A19',
    30: 'LIFX+ BR30',
    31: 'LIFX Z',
    32: 'LIFX Z 2',
    36: 'LIFX Downlight',
    37: 'LIFX Downlight',
    38: 'LIFX Beam',
    43: 'LIFX A19',
    44: 'LIFX BR30',
    45: 'LIFX+ A19',
    46: 'LIFX+ BR30',
    49: 'LIFX Mini',
    50: 'LIFX Mini Day and Dusk',
    51: 'LIFX Mini White',
    52: 'LIFX GU10',
    55: 'LIFX Tile',
    59: 'LIFX Mini Color',
    60: 'LIFX Mini Day and Dusk',
    61: 'LIFX Mini White'}

features_map = {   1: {   'chain': False,
           'color': True,
           'infrared': False,
           'max_kelvin': 9000,
           'min_kelvin': 2500,
           'multizone': False},
    3: {   'chain': False,
           'color': True,
           'infrared': False,
           'max_kelvin': 9000,
           'min_kelvin': 2500,
           'multizone': False},
    10: {   'chain': False,
            'color': False,
            'infrared': False,
            'max_kelvin': 6500,
            'min_kelvin': 2700,
            'multizone': False},
    11: {   'chain': False,
            'color': False,
            'infrared': False,
            'max_kelvin': 6500,
            'min_kelvin': 2700,
            'multizone': False},
    18: {   'chain': False,
            'color': False,
            'infrared': False,
            'max_kelvin': 9000,
            'min_kelvin': 2500,
            'multizone': False},
    20: {   'chain': False,
            'color': True,
            'infrared': False,
            'max_kelvin': 9000,
            'min_kelvin': 2500,
            'multizone': False},
    22: {   'chain': False,
            'color': True,
            'infrared': False,
            'max_kelvin': 9000,
            'min_kelvin': 2500,
            'multizone': False},
    27: {   'chain': False,
            'color': True,
            'infrared': False,
            'max_kelvin': 9000,
            'min_kelvin': 2500,
            'multizone': False},
    28: {   'chain': False,
            'color': True,
            'infrared': False,
            'max_kelvin': 9000,
            'min_kelvin': 2500,
            'multizone': False},
    29: {   'chain': False,
            'color': True,
            'infrared': True,
            'max_kelvin': 9000,
            'min_kelvin': 2500,
            'multizone': False},
    30: {   'chain': False,
            'color': True,
            'infrared': True,
            'max_kelvin': 9000,
            'min_kelvin': 2500,
            'multizone': False},
    31: {   'chain': False,
            'color': True,
            'infrared': False,
            'max_kelvin': 9000,
            'min_kelvin': 2500,
            'multizone': True},
    32: {   'chain': False,
            'color': True,
            'infrared': False,
            'max_kelvin': 9000,
            'min_kelvin': 2500,
            'multizone': True},
    36: {   'chain': False,
            'color': True,
            'infrared': False,
            'max_kelvin': 9000,
            'min_kelvin': 2500,
            'multizone': False},
    37: {   'chain': False,
            'color': True,
            'infrared': False,
            'max_kelvin': 9000,
            'min_kelvin': 2500,
            'multizone': False},
    38: {   'chain': False,
            'color': True,
            'infrared': False,
            'max_kelvin': 9000,
            'min_kelvin': 2500,
            'multizone': True},
    43: {   'chain': False,
            'color': True,
            'infrared': False,
            'max_kelvin': 9000,
            'min_kelvin': 2500,
            'multizone': False},
    44: {   'chain': False,
            'color': True,
            'infrared': False,
            'max_kelvin': 9000,
            'min_kelvin': 2500,
            'multizone': False},
    45: {   'chain': False,
            'color': True,
            'infrared': True,
            'max_kelvin': 9000,
            'min_kelvin': 2500,
            'multizone': False},
    46: {   'chain': False,
            'color': True,
            'infrared': True,
            'max_kelvin': 9000,
            'min_kelvin': 2500,
            'multizone': False},
    49: {   'chain': False,
            'color': True,
            'infrared': False,
            'max_kelvin': 9000,
            'min_kelvin': 2500,
            'multizone': False},
    50: {   'chain': False,
            'color': False,
            'infrared': False,
            'max_kelvin': 4000,
            'min_kelvin': 1500,
            'multizone': False},
    51: {   'chain': False,
            'color': False,
            'infrared': False,
            'max_kelvin': 2700,
            'min_kelvin': 2700,
            'multizone': False},
    52: {   'chain': False,
            'color': True,
            'infrared': False,
            'max_kelvin': 9000,
            'min_kelvin': 2500,
            'multizone': False},
    55: {   'chain': True,
            'color': True,
            'infrared': False,
            'max_kelvin': 9000,
            'min_kelvin': 2500,
            'multizone': False},
    59: {   'chain': False,
            'color': True,
            'infrared': False,
            'max_kelvin': 9000,
            'min_kelvin': 2500,
            'multizone': False},
    60: {   'chain': False,
            'color': False,
            'infrared': False,
            'max_kelvin': 4000,
            'min_kelvin': 1500,
            'multizone': False},
    61: {   'chain': False,
            'color': False,
            'infrared': False,
            'max_kelvin': 2700,
            'min_kelvin': 2700,
            'multizone': False}}

